@@ -3,13 +3,16 @@
 import argparse
 import plotly.express as px
 import pandas as pd
-import datetime
+#import datetime
 import sqlite3
 import time
 import re
 import sys
 from operator import itemgetter
 import shutil
+import numpy as np
+from itertools import product
+from dataclasses import dataclass
 
 from os import listdir
 from dash import Dash, dcc, html, dash_table, Input, Output, State
@@ -22,7 +25,7 @@ AS['database'] = None
 AS['argv'] = None
 AS['start_time'] = None
 
-def update_legend_labels(fig,labels):
+def update_legend_labels(fig, labels):
     for entry in fig.data:
         if entry['name'] in labels:
             entry['name'] = labels[entry['name']]
@@ -81,7 +84,13 @@ def get_databases(directory):
         databases_new.append('%s (%d)' %(databases[index], get_number_of_experiments(directory, databases[index])))
     return databases_new
 
-def run(directory, ip="127.0.0.1", port=8080, x_axis="delay", y_axis="length", aspect_ratio=0, auto_update_interval=0, debug=False):
+@dataclass
+class Heatmap():
+    x_number_of_bins: int = 10
+    y_number_of_bins: int = 10
+    color_scale:str = "findus"
+
+def run(directory, ip="127.0.0.1", port=8080, x_axis="delay", y_axis="length", aspect_ratio=0, auto_update_interval=0, debug=False, heatmap:Heatmap=None):
     DATABASE_DIRECTORY = directory
     app = Dash(__name__, external_stylesheets=[dbc.themes.LUX])
     app.css.config.serve_locally = True
@@ -96,15 +105,18 @@ def run(directory, ip="127.0.0.1", port=8080, x_axis="delay", y_axis="length", a
                     html.Option(value="match_string(response, 'ets')"),
                     html.Option(value="match_hex(response, '661b')"),
                     html.Option(value="color = 'G'"),
-                    html.Option(value=f"length > 100"),
-                    html.Option(value=f"delay > 100"),
-                    html.Option(value=f"voltage = 400"),
+                    html.Option(value="length > 100"),
+                    html.Option(value="delay > 100"),
+                    html.Option(value="voltage = 400"),
                 ]),
                 dcc.Input(id='query-input', type="text", list='examples', style={'width':'80%','display': 'inline-block'}, placeholder="SELECT * FROM experiments WHERE"),
                 dcc.Dropdown(id='graph-dropdown', style={'width':'100%'}),
                 html.Center([
                     dcc.Graph(id='graph', style={'width':'80%'}),
                 ]),
+                html.Center([
+                    dcc.Graph(id='graph_opt', style=({'display': 'none'} if heatmap is None else {'width':'80%'})),
+                ]),
                 html.P('re.search(*, response)'),
                 dcc.Input(id='recolor-green', type="text", placeholder="green", style={'width':'15%'}),
                 dcc.Input(id='recolor-green-label', type="text", placeholder="green-label", style={'width':'10%'}),
@@ -129,7 +141,7 @@ def run(directory, ip="127.0.0.1", port=8080, x_axis="delay", y_axis="length", a
                 dcc.RadioItems(id='combine-data', options=['Yes', 'No'], value='Yes', inline=True),
                 html.Div(id='data',style={'width':'100%', 'height':'80%', 'border-style':'none'}),
                 html.Label('Arguments:'),
-                html.Label('lalala', id='argv'),
+                html.Label('argv_label', id='argv'),
 
                 # timer for update_graph callback
                 dcc.Interval(
@@ -138,8 +150,8 @@ def run(directory, ip="127.0.0.1", port=8080, x_axis="delay", y_axis="length", a
                     n_intervals=0,
                     disabled=(auto_update_interval == 0)
                 )
-            ],style={'width':'80%','border-style':'none','margin':'0 auto'}),
-        ],style={'width':'100%', 'border-style':'none', 'margin-top':'100px','margin-bottom':'100px'})
+            ], style={'width':'80%','border-style':'none','margin':'0 auto'}),
+        ], style={'width':'100%', 'border-style':'none', 'margin-top':'100px','margin-bottom':'100px'})
 
     # callback for database list
     @app.callback(
@@ -185,6 +197,7 @@ def run(directory, ip="127.0.0.1", port=8080, x_axis="delay", y_axis="length", a
     # callback for database selection  
     @app.callback(
         Output('graph','figure'),
+        Output('graph_opt','figure'),
         Output('data','children'),
         Input('update-button', 'n_clicks'),
         Input('graph-dropdown', 'value'),
@@ -231,7 +244,7 @@ def run(directory, ip="127.0.0.1", port=8080, x_axis="delay", y_axis="length", a
 
         # perform the query based on the query extension
         if query is not None and query != '':
-            query = f'SELECT * FROM experiments WHERE %s;' %(query)
+            query = f'SELECT * FROM experiments WHERE {query};'
         else:
             query = 'SELECT * FROM experiments;'
 
@@ -271,23 +284,88 @@ def run(directory, ip="127.0.0.1", port=8080, x_axis="delay", y_axis="length", a
             x = x_axis,
             y = y_axis,
             render_mode = "webgl",
-            color = "color", 
+            color = "color",
             labels = {"color":f"Classification ({nr_of_current_experiments:,})",y_axis:y_axis, x_axis:x_axis},
-            color_discrete_map = { 
+            color_discrete_map = {
                 "G": "green",
-                "Y": "yellow", 
-                "M" : "magenta", 
+                "Y": "yellow",
+                "M" : "magenta",
                 "O": "orange",
                 "C": "cyan",
                 "B": "blue",
                 "Z": "black",
-                "R": "red", 
+                "R": "red",
             },
             category_orders = {
                 "color" : ["G","Y","M","O","C","B","R"]
             },
         )
 
+        fig_opt = None
+        if heatmap is not None:
+            # Assume x_edges and y_edges already created
+            x_edges = np.linspace(df[x_axis].min(), df[x_axis].max(), heatmap.x_number_of_bins + 1, endpoint=True)
+            y_edges = np.linspace(df[y_axis].min(), df[y_axis].max(), heatmap.y_number_of_bins + 1, endpoint=True)
+            # include lower and upper limits
+            x_edges[-1] = df[x_axis].max()
+            y_edges[-1] = df[y_axis].max()
+            x_step = (df[x_axis].max() - df[x_axis].min()) / (heatmap.x_number_of_bins)
+            y_step = (df[y_axis].max() - df[y_axis].min()) / (heatmap.y_number_of_bins)
+            x_min = df[x_axis].min() - x_step
+            y_min = df[y_axis].min() - y_step
+            x_edges = np.insert(x_edges, 0, x_min)
+            y_edges = np.insert(y_edges, 0, y_min)
+
+            # Bin data
+            df["x_bin"] = pd.cut(df[x_axis], bins=x_edges, labels=False, include_lowest=True)
+            df["y_bin"] = pd.cut(df[y_axis], bins=y_edges, labels=False, include_lowest=True)
+
+            # Filter red only
+            red_df = df[df["color"] == "R"]
+            green_df = df[df["color"] == "G"]
+
+            # Count red points in each bin
+            red_counts = red_df.groupby(["x_bin", "y_bin"]).size().reset_index(name="red_count")
+            green_counts = green_df.groupby(["x_bin", "y_bin"]).size().reset_index(name="green_count")
+
+            # merge counts and calculate score
+            heatmap_data = pd.merge(red_counts, green_counts, on=["x_bin", "y_bin"], how="outer").fillna(0)
+            heatmap_data["score"] = heatmap_data["red_count"] / (heatmap_data["red_count"] + heatmap_data["green_count"])
+
+            # Create all possible bin combinations
+            all_bins = pd.DataFrame(list(product(range(heatmap.x_number_of_bins + 1), range(heatmap.y_number_of_bins + 1))), columns=["x_bin", "y_bin"])
+
+            # Merge with actual data to fill missing bins
+            heatmap_data = pd.merge(all_bins, heatmap_data, on=["x_bin", "y_bin"], how="left")
+            heatmap_data["score"] = heatmap_data["score"].fillna(0)
+
+            # Map bin numbers to bin centers for plotting
+            x_centers = 0.5 * (x_edges[:-1] + x_edges[1:])
+            y_centers = 0.5 * (y_edges[:-1] + y_edges[1:])
+            heatmap_data["x"] = heatmap_data["x_bin"].map(dict(enumerate(x_centers)))
+            heatmap_data["y"] = heatmap_data["y_bin"].map(dict(enumerate(y_centers)))
+
+            if heatmap.color_scale == "findus":
+                costum_scale=[
+                    "#FFEECC",  # White
+                    "#FFA500",  # Orange
+                    "#FF0000",  # Red
+                    "#FF00FF"   # Magenta
+                ]
+                heatmap.color_scale = costum_scale
+
+            # Plot heatmap
+            fig_opt = px.density_heatmap(
+                heatmap_data,
+                x="x",
+                y="y",
+                z="score",
+                nbinsx=heatmap.x_number_of_bins + 1,
+                nbinsy=heatmap.y_number_of_bins + 1,
+                color_continuous_scale=heatmap.color_scale,
+                labels={"score": "score", "x": x_axis, "y": y_axis}
+            )
+
         # compute elapsed time
         # elapsed_time_in_seconds = int(time.time()-get_start_time(DATABASE_DIRECTORY, database))
         # elapsed_time = datetime.timedelta(seconds=elapsed_time_in_seconds)
@@ -300,14 +378,58 @@ def run(directory, ip="127.0.0.1", port=8080, x_axis="delay", y_axis="length", a
         else:
             fig.update_layout(title_text=f"{database[:-7]}", title_x=0.5)
 
+        if fig_opt is not None:
+            fig_opt.update_layout(title_text=f"{database[:-7]}", title_x=0.5)
+
+        # calculate the average percentage and the standard deviation for every category
+        # Divide the experiments carried out into sub-experiments and calculate the average 
+        # quotient (xi) for each subset. Also calculate the average quotient of all experiments (mu).
+        # The standard deviation can then be calculated using the following formula, where N is the number of experiments:
+        # sigma = Sqrt(Sum((xi - mu)**2) / N)
+        # Note that this approximation is dependent on the number of subsets the experiments are divided into (samples).
+        # If the sample size is equal to 1, the standard deviation is 0.
+        # If the sample size is large, the standard deviation is big and meaningless.
+        # Both extreme values are not meaningful, which is why a samples size in between must be selected.
+        # Select >= 10 samples for a rough estimate.
+        # Select >= 30 samples for a statistically reliable estimate.
+        # Select >= 100 samples for solid confidence, especially with noisy data.
+        samples = 10
+        bin_size = int(len(df) / samples)
+        if bin_size == 0:
+            bin_size = 1
+        count = {'G':0, 'Y':0, 'M':0, 'O':0, 'C':0, 'B':0, 'Z':0, 'R':0}
+        stddev = {'G':0, 'Y':0, 'M':0, 'O':0, 'C':0, 'B':0, 'Z':0, 'R':0}
+        avg = {'G':0, 'Y':0, 'M':0, 'O':0, 'C':0, 'B':0, 'Z':0, 'R':0}
+        for color in stddev:
+            cnt = len(df.query(f"color == '{color}'"))
+            count[color] = cnt
+            if cnt > 0:
+                mu = cnt / len(df) * 100
+            else:
+                mu = 0.0
+            avg[color] = mu
+            
+            avg_per_bin = []
+            for bins in range(0, int(len(df) / bin_size) - 1):
+                subset = df[bin_size * bins:bin_size * (bins + 1)]
+                cnt_subset = len(subset.query(f"color == '{color}'"))
+                if cnt_subset > 0 and len(subset) > 0:
+                    mu_subset = cnt_subset / len(subset)
+                else:
+                    mu_subset = 0.0
+                avg_per_bin.append(mu_subset)
+            sigma = 0
+            if cnt > 0:
+                for xi in avg_per_bin:
+                    sigma += (xi - mu)**2 / cnt
+                sigma = sigma**0.5
+            stddev[color] = sigma
+        #print(stddev)
+
+
         # update labels in legenda
         def make_label(color, label, df):
-            count = len(df.query(f"color == '{color}'"))
-            if count > 0:
-                percentage = "{:.1%}".format(count/len(df))
-            else:
-                percentage = "{:.1%}".format(0)
-            return { color: f'{label} ( {count} / {percentage} )'}
+            return { color: f'{label} ( {count[color]} / {avg[color]:.1f} % +- {stddev[color]:.1f} % )'}
 
         labels = {}
         labels.update(make_label('G', greenlabel, df))
@@ -327,10 +449,8 @@ def run(directory, ip="127.0.0.1", port=8080, x_axis="delay", y_axis="length", a
 
         # TODO: rewrite this code
         if combine == 'Yes':
-
             combined_records = []
             for record in records:
-
                 # decode response to make sure it's compatible with json
                 response_hex = record['response'].hex(' ')
                 response = record['response'].decode('utf-8', errors='replace')
@@ -382,7 +502,6 @@ def run(directory, ip="127.0.0.1", port=8080, x_axis="delay", y_axis="length", a
  
         else:
             columns = ['id', 'color', y_axis, x_axis, 'rlen', 'response','response_hex']
-
             all_records = []
 
             for record in records:
@@ -436,12 +555,12 @@ def run(directory, ip="127.0.0.1", port=8080, x_axis="delay", y_axis="length", a
 
         if debug:
             done = round(time.time() * 1000)
-            print('It took %d miliseconds to generate this data.' %(done - now))
+            print('It took %d milliseconds to generate this data.' %(done - now))
 
-        return fig,data
+        return fig, fig_opt, data
 
     # start server on localhost
-    app.run_server(host=ip, port=port, debug=True)
+    app.run(host=ip, port=port, debug=True)
 
 def main(argv=sys.argv):
     __version__ = "0.1"
@@ -455,11 +574,20 @@ def main(argv=sys.argv):
     parser.add_argument("--ip", help="Server address", required=False, default='127.0.0.1')
     parser.add_argument("-x", help="parameter to plot on the x-axis", required=False, default='delay')
     parser.add_argument("-y", help="parameter to plot on the y-axis", required=False, default='length')
-    parser.add_argument("--aspect-ratio", help="aspect ratio of the plot relative to x-axis", required=False, default=0.38, type=float)
-    parser.add_argument("--auto-update", help="Whether to update the plot automatically. Optionally pass the update interval in seconds.", required=False, default=0, type=int, nargs='?', const=1)
+    parser.add_argument("--aspect-ratio", help="aspect ratio of the plot relative to x-axis", required=False, default=0, type=float)
+    parser.add_argument("--auto-update", help="Whether to update the plot automatically. Optionally pass the update interval in seconds.", required=False, type=int, nargs='?', const=1, default=0)
+    parser.add_argument("--heatmap", action="store_true", help="Generate a heat map", required=False)
+    parser.add_argument("--x-number-of-bins", "--x-bins", help="Number of bins of the x-axis for the heat map", required=False, default=10, type=int)
+    parser.add_argument("--y-number-of-bins", "--y-bins", help="Number of bins of the y-axis for the heat map", required=False, default=10, type=int)
+    parser.add_argument("--color-scale", help="Color scale to use for the heat map (findus, Blues, Reds, Greys, PuRd, YlOrRd).", required=False, default="findus", type=str)
 
     args = parser.parse_args()
-    run(directory=args.directory, ip=args.ip, port=args.port, x_axis=args.x, y_axis=args.y, aspect_ratio=args.aspect_ratio, auto_update_interval=args.auto_update, debug=True)
+
+    heatmap = None
+    if args.heatmap:
+        heatmap = Heatmap(x_number_of_bins=args.x_number_of_bins, y_number_of_bins=args.y_number_of_bins, color_scale=args.color_scale)
+
+    run(directory=args.directory, ip=args.ip, port=args.port, x_axis=args.x, y_axis=args.y, aspect_ratio=args.aspect_ratio, auto_update_interval=args.auto_update, heatmap=heatmap, debug=True)
 
 if __name__ == "__main__":
     main()
